--	NEW Player class colors.
local function whoaUnitClass(healthbar, unit)
	if UnitIsPlayer(unit) and UnitIsConnected(unit) and UnitClass(unit) and (cfg.classColor) then
		_, class = UnitClass(unit);
		local c = RAID_CLASS_COLORS[class];
		healthbar:SetStatusBarColor(c.r, c.g, c.b);
	elseif UnitIsPlayer(unit) and (not UnitIsConnected(unit)) then
		healthbar:SetStatusBarColor(0.5,0.5,0.5);
	else
		healthbar:SetStatusBarColor(0,0.9,0);
	end
end
hooksecurefunc("UnitFrameHealthBar_Update", whoaUnitClass)
hooksecurefunc("HealthBar_OnValueChanged", function(self)
	whoaUnitClass(self, self.unit)
end)

-- Override: Enemy player class color set to red.
local function playerHPColorOverride(healthbar, unit)
	if (cfg.reactColor == true) then
		if (UnitIsPlayer(unit) or UnitPlayerControlled(unit)) and UnitIsEnemy("player",unit) then
			healthbar:SetStatusBarColor(0.8,0.0,0.0);	-- red color.
		end
	end
end
-- hooksecurefunc("UnitFrameHealthBar_Update", playerHPColorOverride)
-- hooksecurefunc("HealthBar_OnValueChanged", function(self)
-- playerHPColorOverride(self, self.unit)
-- end)

	--	Blizzard´s target unit reactions HP color
function npcReactionBrightColors()
	if cfg.blizzColors == true then
		FACTION_BAR_COLORS = {
			[1] = {r =  0.8, g = 0.0, b = 0.0},
			[2] = {r =  0.8, g = 0.0, b = 0.0},
			[3] = {r =  0.8, g = 0.0, b = 0.0},
			[4] = {r =  1, g =  0.8, b = 0.0},
			[5] = {r = 0.0, g = 0.8, b = 0.0},
			[6] = {r = 0.0, g = 0.8, b = 0.0},
			[7] = {r = 0.0, g = 0.8, b = 0.0},
			[8] = {r = 0.0, g = 0.8, b = 0.0},
		};
	end
end
hooksecurefunc("TargetFrame_CheckFaction", npcReactionBrightColors)

--	Whoa´s customs target unit reactions HP colors.
local function npcReactionColors(healthbar, unit)
	if UnitExists(unit) and not UnitIsPlayer(unit) and UnitIsTapDenied(unit) then
		healthbar:SetStatusBarColor(0.5, 0.5, 0.5);
	elseif UnitExists(unit) and not UnitIsPlayer(unit) and cfg.reactColor then
		local reaction = FACTION_BAR_COLORS[UnitReaction(unit,"player")];
		if reaction then
			healthbar:SetStatusBarColor(reaction.r, reaction.g, reaction.b);
		elseif not reaction and cfg.blizzColors then
			healthbar:SetStatusBarColor(0,0.6,0.1);
		elseif not reaction then
			healthbar:SetStatusBarColor(0,0.6,0.1);
		end
	end
	-- Override: Player set enemy player class color to red.
	-- if (UnitIsPlayer(unit) or UnitPlayerControlled(unit)) and UnitIsEnemy("player",unit) then
			-- healthbar:SetStatusBarColor(0.8,0.0,0.0);	-- red color.
	-- end
end
hooksecurefunc("UnitFrameHealthBar_Update", npcReactionColors)
hooksecurefunc("HealthBar_OnValueChanged", function(self)
	npcReactionColors(self, self.unit)
end)

-- aura positioning constants
local AURA_START_X = 5;
local AURA_START_Y = 32;
local AURA_OFFSET_Y = 3;
local LARGE_AURA_SIZE = cfg.largeAuraSize;	-- Default 21.
local SMALL_AURA_SIZE = cfg.smallAuraSize;	-- Default 17.
local AURA_ROW_WIDTH = 130;
local TOT_AURA_ROW_WIDTH = 101;
local NUM_TOT_AURA_ROWS = 2;	-- TODO: replace with TOT_AURA_ROW_HEIGHT functionality if this becomes a problem

function whoaUpdateAuraPositions(self, auraName, numAuras, numOppositeAuras, largeAuraList, updateFunc, maxRowWidth, offsetX, mirrorAurasVertically)
	local size;
	local offsetY = AURA_OFFSET_Y;
	local rowWidth = 0;
	local firstBuffOnRow = 1;
	for i=1, numAuras do
		if ( largeAuraList[i] ) then
			size = cfg.largeAuraSize;
			offsetY = AURA_OFFSET_Y + AURA_OFFSET_Y;
		else
			size = cfg.smallAuraSize;
		end
		if ( i == 1 ) then
			rowWidth = size;
			self.auraRows = self.auraRows + 1;
		else
			rowWidth = rowWidth + size + offsetX;
		end
		if ( rowWidth > maxRowWidth ) then
			updateFunc(self, auraName, i, numOppositeAuras, firstBuffOnRow, size, offsetX, offsetY, mirrorAurasVertically);

			rowWidth = size;
			self.auraRows = self.auraRows + 1;
			firstBuffOnRow = i;
			offsetY = AURA_OFFSET_Y;
			if ( self.auraRows > NUM_TOT_AURA_ROWS ) then
				maxRowWidth = AURA_ROW_WIDTH;
			end
		else
			updateFunc(self, auraName, i, numOppositeAuras, i - 1, size, offsetX, offsetY, mirrorAurasVertically);
		end
	end
end
hooksecurefunc("TargetFrame_UpdateAuraPositions", whoaUpdateAuraPositions)

function UpdateBuffAnchor(self, buffName, index, numDebuffs, anchorIndex, size, offsetX, offsetY, mirrorVertically)
	local point, relativePoint;
	local startY, auraOffsetY;
	if ( mirrorVertically ) then
		point = "BOTTOM";
		relativePoint = "TOP";
		startY = -5;
		if ( self.threatNumericIndicator:IsShown() ) then
			startY = startY + self.threatNumericIndicator:GetHeight();
		end
		offsetY = - offsetY;
		auraOffsetY = -AURA_OFFSET_Y;
	else
		point = "TOP";
		relativePoint="BOTTOM";
		startY = AURA_START_Y - 3;
		auraOffsetY = AURA_OFFSET_Y;
	end
	local buff = _G[buffName..index];
	if ( index == 1 ) then
		if ( UnitIsFriend("player", self.unit) or numDebuffs == 0 ) then
			buff:SetPoint(point.."LEFT", self, relativePoint.."LEFT", AURA_START_X, startY);
		else
			buff:SetPoint(point.."LEFT", self.debuffs, relativePoint.."LEFT", 0, -offsetY);
		end
		self.buffs:SetPoint(point.."LEFT", buff, point.."LEFT", 0, 0);
		self.buffs:SetPoint(relativePoint.."LEFT", buff, relativePoint.."LEFT", 0, -auraOffsetY);
		self.spellbarAnchor = buff;
	elseif ( anchorIndex ~= (index-1) ) then
		buff:SetPoint(point.."LEFT", _G[buffName..anchorIndex], relativePoint.."LEFT", 0, -offsetY);
		self.buffs:SetPoint(relativePoint.."LEFT", buff, relativePoint.."LEFT", 0, -auraOffsetY);
		self.spellbarAnchor = buff;
	else
		buff:SetPoint(point.."LEFT", _G[buffName..anchorIndex], point.."RIGHT", offsetX, 0);
	end
	buff:SetWidth(size);
	buff:SetHeight(size);
end
hooksecurefunc("TargetFrame_UpdateBuffAnchor", UpdateBuffAnchor)

function UpdateDebuffAnchor(self, debuffName, index, numBuffs, anchorIndex, size, offsetX, offsetY, mirrorVertically)
	local buff = _G[debuffName..index];
	local isFriend = UnitIsFriend("player", self.unit);
	local point, relativePoint;
	local startY, auraOffsetY;
	if ( mirrorVertically ) then
		point = "BOTTOM";
		relativePoint = "TOP";
		startY = -5;
		if ( self.threatNumericIndicator:IsShown() ) then
			startY = startY + self.threatNumericIndicator:GetHeight();
		end
		offsetY = - offsetY;
		auraOffsetY = -AURA_OFFSET_Y;
	else
		point = "TOP";
		relativePoint="BOTTOM";
		startY = AURA_START_Y - 3;
		auraOffsetY = AURA_OFFSET_Y;
	end
	if ( index == 1 ) then
		if ( isFriend and numBuffs > 0 ) then
			buff:SetPoint(point.."LEFT", self.buffs, relativePoint.."LEFT", 0, -offsetY);
		else
			buff:SetPoint(point.."LEFT", self, relativePoint.."LEFT", AURA_START_X, startY);
		end
		self.debuffs:SetPoint(point.."LEFT", buff, point.."LEFT", 0, 0);
		self.debuffs:SetPoint(relativePoint.."LEFT", buff, relativePoint.."LEFT", 0, -auraOffsetY);
		if ( ( isFriend ) or ( not isFriend and numBuffs == 0) ) then
			self.spellbarAnchor = buff;
		end
	elseif ( anchorIndex ~= (index-1) ) then
		buff:SetPoint(point.."LEFT", _G[debuffName..anchorIndex], relativePoint.."LEFT", 0, -offsetY);
		self.debuffs:SetPoint(relativePoint.."LEFT", buff, relativePoint.."LEFT", 0, -auraOffsetY);
		if ( ( isFriend ) or ( not isFriend and numBuffs == 0) ) then
			self.spellbarAnchor = buff;
		end
	else
		buff:SetPoint(point.."LEFT", _G[debuffName..(index-1)], point.."RIGHT", offsetX, 0);
	end
	buff:SetWidth(size);
	buff:SetHeight(size);
	local debuffFrame =_G[debuffName..index.."Border"];
	debuffFrame:SetWidth(size+2);
	debuffFrame:SetHeight(size+2);
end
hooksecurefunc("TargetFrame_UpdateDebuffAnchor", UpdateDebuffAnchor)

Target_Spellbar_AdjustPosition = function(self) 
	local parentFrame = self:GetParent();
	if ( self.boss ) then
		self:SetPoint("TOPLEFT", parentFrame, "BOTTOMLEFT", 25, 10 );
	elseif ( parentFrame.haveToT ) then	-- if have tot.
		if ( parentFrame.buffsOnTop or parentFrame.auraRows <= 1 ) then
			self:SetPoint("TOPLEFT", parentFrame, "BOTTOMLEFT", 25, -21 );
		else
			self:SetPoint("TOPLEFT", parentFrame.spellbarAnchor, "BOTTOMLEFT", 20, -28);
		end
	elseif ( parentFrame.haveElite ) then
		if ( parentFrame.buffsOnTop or parentFrame.auraRows <= 1 ) then
			self:SetPoint("TOPLEFT", parentFrame, "BOTTOMLEFT", 25, -21 );
		else
			self:SetPoint("TOPLEFT", parentFrame.spellbarAnchor, "BOTTOMLEFT", 20, -28);
		end
	else
		if ( (not parentFrame.buffsOnTop) and parentFrame.auraRows > 0 ) then
			self:SetPoint("TOPLEFT", parentFrame.spellbarAnchor, "BOTTOMLEFT", 20, -20);
		else
			self:SetPoint("TOPLEFT", parentFrame, "BOTTOMLEFT", 25, 3 );	-- target no tot
		end
	end
end
TargetFrameSpellBar:SetScript("OnShow", nil)
FocusFrameSpellBar:SetScript("OnShow", nil)
-- hooksecurefunc("TargetFrame_UpdateBuffAnchor", function(_, name, i) _G[name..i]:SetSize(28, 28) end)
-- hooksecurefunc("TargetFrame_UpdateDebuffAnchor", function(_, name, i) _G[name..i]:SetSize(25, 25) end)

-- Unit frames Status text modifier and reformat.
local function customStatusTex(statusFrame, textString, value, valueMin, valueMax)
	local xpValue = UnitXP("player");
	local xpMaxValue = UnitXPMax("player");
	
	if( statusFrame.LeftText and statusFrame.RightText ) then
		statusFrame.LeftText:SetText("");
		statusFrame.RightText:SetText("");
		statusFrame.LeftText:Hide();
		statusFrame.RightText:Hide();
	end
	
	local k,m=1e3
	local m=k*k
	
	--	whoa custom status text abbreviation.	
	valueDisplay	=	((value>1e3 and value<1e5 and format("%1.3f",value/k)) or 
						(value>=1e5 and value<1e6 and format("%1.0f K",value/k)) or 
						(value>=1e6 and value<1e9 and format("%1.2f M",value/m)) or 
						(value>=1e9 and format("%1.1f M",value/m)) or value )
						
	valueMaxDisplay	=	((valueMax>1e3 and valueMax<1e5 and format("%1.3f",valueMax/k)) or
						(valueMax>=1e5 and valueMax<1e6 and format("%1.0f K",valueMax/k)) or 
						(valueMax>=1e6 and valueMax<1e9 and format("%1.2f M",valueMax/m)) or 
						(valueMax>=1e9 and format("%1.1f M",valueMax/m)) or valueMax )

	--	Blizzards default status text abbreviation.							
	-- valueDisplay = AbbreviateLargeNumbers(value);
	-- valueMaxDisplay = AbbreviateLargeNumbers(valueMax);
	-- xpValueDisplay = BreakUpLargeNumbers(xpValue);
	-- xpMaxValueDisplay = BreakUpLargeNumbers(xpMaxValue);						
	
	if ( ( tonumber(valueMax) ~= valueMax or valueMax > 0 ) and not ( statusFrame.pauseUpdates ) ) then
		statusFrame:Show();
		
		if ( (statusFrame.cvar and GetCVar(statusFrame.cvar) == "1" and statusFrame.textLockable) or statusFrame.forceShow ) then
			textString:Show();
		elseif ( statusFrame.lockShow > 0 and (not statusFrame.forceHideText) ) then
			textString:Show();
		else
			textString:SetText("");
			textString:Hide();
			return;
		end
		local textDisplay = GetCVar("statusTextDisplay");
		if ( value and valueMax > 0 and ( (textDisplay ~= "NUMERIC" and textDisplay ~= "NONE") or statusFrame.showPercentage ) and not statusFrame.showNumeric) then
			if ( value == 0 and statusFrame.zeroText ) then
				textString:SetText(statusFrame.zeroText);
				statusFrame.isZero = 1;
				textString:Show();
			elseif ( textDisplay == "BOTH" and not statusFrame.showPercentage) then
				if( statusFrame.LeftText and statusFrame.RightText ) then
					if(not statusFrame.powerToken or statusFrame.powerToken == "MANA") then
						statusFrame.LeftText:SetText(math.ceil((value / valueMax) * 100) .. "%");
						if value == 0 then statusFrame.LeftText:SetText(""); end
						statusFrame.LeftText:Show();
					end
					statusFrame.RightText:SetText(valueDisplay);
					if value == 0 then statusFrame.RightText:SetText(""); end
					statusFrame.RightText:Show();
					textString:Hide();
				else
					valueDisplay = "(" .. math.ceil((value / valueMax) * 100) .. "%) " .. valueDisplay .. " / " .. valueMaxDisplay;
					if value == 0 then textString:SetText(""); end
				end
				textString:SetText(valueDisplay);
			else
				valueDisplay = math.ceil((value / valueMax) * 100) .. "%";
				if ( statusFrame.prefix and (statusFrame.alwaysPrefix or not (statusFrame.cvar and GetCVar(statusFrame.cvar) == "1" and statusFrame.textLockable) ) ) then
					textString:SetText(statusFrame.prefix .. " " .. valueDisplay);
				else
					textString:SetText(valueDisplay);
				end
				if value == 0 then textString:SetText(""); end
			end
		elseif ( value == 0 and statusFrame.zeroText ) then
			textString:SetText(statusFrame.zeroText);
			statusFrame.isZero = 1;
			textString:Show();
			return;
		else
			statusFrame.isZero = nil;
			if ( statusFrame.prefix and (statusFrame.alwaysPrefix or not (statusFrame.cvar and GetCVar(statusFrame.cvar) == "1" and statusFrame.textLockable) ) ) then
				textString:SetText(statusFrame.prefix.." "..valueDisplay.." / "..valueMaxDisplay)
			else
				textString:SetText(valueDisplay.." / "..valueMaxDisplay);
			end
			if value == 0 then textString:SetText("") end
		end
	else
		textString:Hide();
		textString:SetText("");
		if ( not statusFrame.alwaysShow ) then
			statusFrame:Hide();
		else
			statusFrame:SetValue(0);
		end
	end
end
hooksecurefunc("TextStatusBar_UpdateTextStringWithValues",customStatusTex)

-- Disable/enable unit frames click interaction.
function disableFramesClick()
	if (cfg.noClickFrame == true) then
		for i, v in pairs({	PlayerFrame, PetFrame, TargetFrame, FocusFrame }) do v:SetMouseClickEnabled(false); end
	elseif (cfg.noClickFrame == false) then
		for i, v in pairs({	PlayerFrame, PetFrame, TargetFrame, FocusFrame }) do v:SetMouseClickEnabled(true); end
	end
end

-- Text builder factory.
local function CreateText(name, parentName, parent, point, x, y)
	local fontString = parent:CreateFontString(parentName..name, nil, "GameFontNormalSmall")
	fontString:SetPoint(point, parent, point, x, y)
	return fontString
end

-- Create text frames for dead and ghost.
function createDeadTextFrames()
	whoaPlayerFrameDeadText = CreateText("DeadText", "PlayerFrame", PlayerFrameHealthBar, "CENTER", 0, 0);
	whoaPlayerFrameGhostText = CreateText("GhostText", "PlayerFrame", PlayerFrameHealthBar, "CENTER", 0, 0);
	whoaPlayerFrameDeadText:SetText(DEAD);
	whoaPlayerFrameGhostText:SetText("Ghost");
	
	whoaTargetFrameGhostText = CreateText("GhostText", "TargetFrame", TargetFrameHealthBar, "CENTER", 0, 0);
	whoaTargetFrameGhostText:SetText("Ghost");
	whoaTargetFrameOfflineText = CreateText("OfflineText", "TargetFrame", TargetFrameHealthBar, "CENTER", 0, 0);
	whoaTargetFrameOfflineText:SetText("Offline");
	
	whoaFocusFrameGhostText = CreateText("GhostText", "FocusFrame", FocusFrameHealthBar, "CENTER", 0, 0);
	whoaFocusFrameGhostText:SetText("Ghost");
	whoaFocusFrameOfflineText = CreateText("OfflineText", "FocusFrame", FocusFrameHealthBar, "CENTER", 0, 0);
	whoaFocusFrameOfflineText:SetText("Offline");
end

--	Player dead text switch.
function playerDeadText()
	if UnitIsDead("player") then
		whoaPlayerFrameDeadText:Show();
		whoaPlayerFrameGhostText:Hide();
		PlayerFrameAlternateManaBar:Hide();
	elseif UnitIsGhost("player") then
		whoaPlayerFrameDeadText:Hide();
		whoaPlayerFrameGhostText:Show();
		PlayerFrameAlternateManaBar:Hide();
	else
		whoaPlayerFrameDeadText:Hide();
		whoaPlayerFrameGhostText:Hide();
		PlayerFrameAlternateManaBar:Show();
	end
	if UnitExists("player") and UnitIsDead("player") or UnitIsGhost("player") then
		for i, v in pairs({	PlayerFrameHealthBar.LeftText, PlayerFrameHealthBar.RightText, PlayerFrameManaBar.LeftText, PlayerFrameManaBar.RightText, PlayerFrameTextureFrameManaBarText, PlayerFrameManaBar }) do v:SetAlpha(0); end
	else
		for i, v in pairs({	PlayerFrameHealthBar.LeftText, PlayerFrameHealthBar.RightText, PlayerFrameManaBar.LeftText, PlayerFrameManaBar.RightText, PlayerFrameTextureFrameManaBarText, PlayerFrameManaBar }) do v:SetAlpha(1); end
	end
end

-- Target frame dead text switch.
function targetDeadText()
	if UnitIsDead("target") then
		whoaTargetFrameGhostText:Hide();
		whoaTargetFrameOfflineText:Hide();
	elseif UnitIsGhost("target") then
		whoaTargetFrameGhostText:Show();
		whoaTargetFrameOfflineText:Hide();
	elseif UnitIsPlayer("target") and not UnitIsConnected("target") then
		whoaTargetFrameGhostText:Hide();
		whoaTargetFrameOfflineText:Show();
	else
		whoaTargetFrameGhostText:Hide();
		whoaTargetFrameOfflineText:Hide();
	end
	if UnitExists("target") and UnitIsDead("target") or UnitIsGhost("target") or not UnitIsConnected("target") then
		for i, v in pairs({	TargetFrameHealthBar.LeftText, TargetFrameHealthBar.RightText, TargetFrameManaBar.LeftText, TargetFrameManaBar.RightText, TargetFrameTextureFrameManaBarText, TargetFrameManaBar }) do v:SetAlpha(0); end
	else
		for i, v in pairs({	TargetFrameHealthBar.LeftText, TargetFrameHealthBar.RightText, TargetFrameManaBar.LeftText, TargetFrameManaBar.RightText, TargetFrameTextureFrameManaBarText, TargetFrameManaBar }) do v:SetAlpha(1); end
	end
end

-- Focus frame dead text switch.
function focusDeadText()
	if UnitIsDead("focus") then
		whoaFocusFrameGhostText:Hide();
		whoaFocusFrameOfflineText:Hide();
	elseif UnitIsGhost("focus") then
		whoaFocusFrameGhostText:Show();
		whoaFocusFrameOfflineText:Hide();
	elseif UnitIsPlayer("focus") and not UnitIsConnected("focus") then
		whoaFocusFrameGhostText:Hide();
		whoaFocusFrameOfflineText:Show();
	else
		whoaFocusFrameGhostText:Hide();
		whoaFocusFrameOfflineText:Hide();
	end
	if UnitExists("focus") and UnitIsDead("focus") or UnitIsGhost("focus") or not UnitIsConnected("focus") then
		for i, v in pairs({	FocusFrameHealthBar.LeftText, FocusFrameHealthBar.RightText, FocusFrameManaBar.LeftText, FocusFrameManaBar.RightText, FocusFrameTextureFrameManaBarText, FocusFrameManaBar }) do v:SetAlpha(0); end
	else
		for i, v in pairs({	FocusFrameHealthBar.LeftText, FocusFrameHealthBar.RightText, FocusFrameManaBar.LeftText, FocusFrameManaBar.RightText, FocusFrameTextureFrameManaBarText, FocusFrameManaBar }) do v:SetAlpha(1); end
	end
end

-- Events triggers.
function whoa:Init(event, addon, ...)
	if (event == "PLAYER_ENTERING_WORLD") then
		createDeadTextFrames()
		playerDeadText ()
		disableFramesClick()
	elseif (event == "PLAYER_TARGET_CHANGED") then
		targetDeadText()
		focusDeadText()
	elseif (event == "PLAYER_DEAD") then
		targetDeadText()
		focusDeadText()
		playerDeadText()
	elseif (event == "PLAYER_UNGHOST") then
		playerDeadText()
		targetDeadText()
		focusDeadText()
	elseif (event == "PLAYER_ALIVE") then
		playerDeadText()
		targetDeadText()
		focusDeadText()
	elseif (event == "PLAYER_FOCUS_CHANGED") then
		focusDeadText()
	elseif (event == "PARTY_MEMBER_DISABLE") then
		targetDeadText()
		focusDeadText()
	end

end
-- Events listener.
local whoaEvent = CreateFrame("Frame", "whoaEvent", UIParent)
whoaEvent:RegisterEvent("PLAYER_ENTERING_WORLD");
whoaEvent:RegisterEvent("PLAYER_TARGET_CHANGED")
whoaEvent:RegisterEvent("PLAYER_DEAD")
whoaEvent:RegisterEvent("PLAYER_UNGHOST")
whoaEvent:RegisterEvent("PLAYER_ALIVE")
whoaEvent:RegisterEvent("PLAYER_FOCUS_CHANGED")
whoaEvent:RegisterEvent("PARTY_MEMBER_DISABLE")
whoaEvent:SetScript("OnEvent", whoa.Init)
